import {
  parseCountryData,
  parseMapData,
  getDataExtent,
} from './dataProcessing.js';

import {
  redrawXAxis,
  redrawYAxis,
  toggleAxisScale,
} from './axisStuff.js';

import {
  width,
  height,
  mapWidth,
  plotWidth,
  margin,
  svg,
  svg2,
  tooltip,
} from './svgGlobals.js';

// Here are the names of the attributes I'll pull out of my data
// objects for my various visual channels
// It's useful to declare them once, since I will need to refer to them
// in several places (e.g. setting visual properties, axes, and labels)
var mapColorAttr = 'Urban_rate';
var mapColorScale = d3.scaleLinear();

var plotColorAttr = 'Continent';
var xAttr = 'Life expectancy';
var yAttr = 'Birth Rate';
var rAttr = 'Population';

const plotColorScale = d3.scaleOrdinal(d3.schemeCategory10);

var xExtent = undefined;
var yExtent = undefined;
var xScale = d3.scaleLinear();
var yScale = d3.scaleLinear();
var xScaleType = 'linear';
var yScaleType = 'linear';

var allCountryData;
var allMapData;

// Since we need to do a little error checking here, lets write a
// nice function so that our attribute setting looks cleaner
function getCountryColor(d) {
  // First, try to get the data for a given country
  let entry = allCountryData.get(d.properties.name);
  // And check if we actually retrieved a value; if not, we'll only
  // get "undefined" for missing data and can use a default 'grey' color
  return entry == undefined
    ? 'grey'
    : mapColorScale(entry[mapColorAttr]);
}

// This one is way more straightforward, since I assume I always have
// complete data for each point on the scatter plot
function getPointColor(d) {
  return plotColorScale(d[plotColorAttr]);
}

// Here's the main rendering code; whenever I call this function, it
// will draw a choropleth map from the mapData and a scatter plot from
// the countryData
function render(mapData, countryData) {
  // Here's how I'm going to convert lat/long coordinates into pixel values
  // on the screen; I can use whichever projection I like best
  var projection = d3
    .geoEqualEarth()
    .fitSize([mapWidth, height], mapData);
  // And here's the part that converts a sequence of points around a
  // polygon into a valid "d" attribute for a <path> tag
  var path = d3.geoPath(projection);

  // Here's a scale that converts attribute values into a color that is
  // on a linear scale between white and blue
  mapColorScale = d3
    // The type of scale: linear, log, band, ordinal, etc.
    .scaleLinear()
    // The range of possible input values:
    .domain(
      getDataExtent(countryData.values(), mapColorAttr),
    )
    // And the range of possible output values:
    .range(['white', 'blue']);

  // Here are the rest of the scaling functions (for the scatter plot)
  // We'll need to handle the x/y positions, radius, and color

  // Boring linear scale for x
  xExtent = getDataExtent(countryData.values(), xAttr);
  xScale = d3
    .scaleLinear()
    .domain(xExtent)
    .range([margin, plotWidth - margin]);

  // Boring linear scale for y
  yExtent = getDataExtent(countryData.values(), yAttr);
  yScale = d3
    .scaleLinear()
    .domain(yExtent)
    // Have you noticed that the range for the y axis is backwards?
    // It maps the lowest input to the highest pixel value, and the highest
    // input to 0 pixels. This is because the origin (0,0) for graphics is
    // in the top left, instead of the bottom left like mathematicians might
    // expect
    .range([height - margin, margin]);

  const rScale = d3
    .scaleLinear()
    .domain(getDataExtent(countryData.values(), rAttr))
    .range([3, 10]);

  // Ok, down to brass tacks: let's add <path> elements to our first svg
  // that encode the borders of each country. I'll save this selection as
  // a variable, so that I can modify the countries later if I need to
  const countryMarks = svg
    // I like to append a <g> element that will hold all my paths, just so
    // I have the option of easily adjusting or moving the entire group
    .append('g')
    // Since the svg is originally empty (I just created it, after all), any
    // selection here will be empty; its only purpose is to give me a place
    // to join my data
    .selectAll('path')
    // Once I have the (empty) selection, I will join all of my data points to
    // it. I know that the structure of geojson is for there to be a "features"
    // attribute that contains the Array of all the geo objects, and I want to
    // bind each geo object to its own path
    .data(mapData.features)
    .join('path')
    // Ok, now that I have a shiny new <path> element with a single country's
    // geo data bound to it, I'm ready to start setting them up.
    // First, the fill color: I'll use the function I wrote above that converts
    // a country name string into a scaled color value
    .attr('fill', getCountryColor)
    // Next, I'll set the "id" attribute, which will give me the ability to
    // select linked items that have the same id
    .attr('id', (d) => d.properties.name)
    .classed('poly', true)
    // Finally, lets use the default geojson setup of setting the 'd' attribute
    // using our path generator function. This is sort of magic that works under
    // the hood
    .attr('d', path);

  redrawXAxis(0, xScale, xAttr);
  redrawYAxis(0, yScale, yAttr);

  // Now we need to add points to our scatter plot; again I will save the
  // selection as a variable so I can modify it more later
  const scatterMarks = svg2
    .append('g')
    .selectAll('circle')
    .data(countryData.values())
    .join('circle')
    .attr('r', (d) => rScale(d[rAttr]))
    .attr('fill', getPointColor)
    .attr('id', (d) => d.Country)
    .attr('class', (d) => `dot ${d.Continent}`)
    .style('opacity', 0.4)
    .attr('cx', (d) => xScale(d[xAttr]))
    .attr('cy', (d) => yScale(d[yAttr]));

  // Okay! It's later now: let's hook up the mouse events for our countries
  // and points in the scatter plot

  // Planning ahead, the first thing I'm going to want to do is highlight
  // matching marks in the linked plots, and I know they're linked by their
  // "id" attribute matching, and differentiated by having different classes
  // ("poly" for the map marks, and "dot" for the scatterplot marks)
  function highlightMarks(id) {
    // Select everything with this id (polygon and dot), and set the fill
    // color for both
    d3.selectAll(`#${id}.poly`)
      .attr('fill', 'red')
      .style('opacity', 1);
    // Select just the dot with the right id, and set its radius
    d3.selectAll(`#${id}.dot`)
      .attr('r', 8)
      .style('opacity', 1)
      .style('stroke', 'black');
  }

  // Finally, I can use my scaling functions for color and radius to reset
  // back to normal when I mouse out
  function resetMarks(id) {
    d3.selectAll(`#${id}.dot`)
      .attr('fill', getPointColor)
      .attr('r', (d) => rScale(d[rAttr]))
      .style('opacity', 0.4)
      .style('stroke', 'none');
    d3.selectAll(`#${id}.poly`)
      .transition()
      .duration(600)
      .attr('fill', getCountryColor);
  }

  // I think it would be useful to tap into a popout effect to visually
  // distinguish marks that are linked to the same continent
  function highlightContinent(continent) {
    d3.selectAll(`.${continent}.dot`)
      .style('stroke', 'black')
      .style('opacity', 1);
  }

  function unHighlightContinent(continent) {
    d3.selectAll(`.${continent}.dot`)
      .style('stroke', 'none')
      .style('opacity', 0.4);
  }

  function popUpTooltip(event, data) {
    let country = countryData.get(data.properties.name);
    tooltip
      .style('opacity', 1)
      .style('z-index', 10)
      .html(
        `${country.Country_orig}<br/>population ${country.Population_orig}`,
      );
  }

  function adjustTooltip(event) {
    tooltip
      .style('left', event.x - 80 + 'px')
      .style('top', event.y - 80 + 'px');
  }

  function hideTooltip() {
    tooltip.style('opacity', 0).style('z-index', -1);
  }

  function addCentroid(continent) {
    let validCountries = [...countryData.values()].filter(
      (d) => d.Continent == continent,
    );
    let xVal = d3.mean(
      d3.map(validCountries, (d) => d[xAttr]),
    );
    let yVal = d3.mean(
      d3.map(validCountries, (d) => d[yAttr]),
    );
    svg2
      .append('circle')
      .attr('id', 'mean')
      .attr('cx', xScale(xVal))
      .attr('cy', height - margin)
      .attr('fill', 'black')
      .attr('r', 5);
    svg2
      .append('circle')
      .attr('id', 'mean')
      .attr('cx', margin)
      .attr('cy', yScale(yVal))
      .attr('fill', 'black')
      .attr('r', 5);
  }

  function removeCentroid() {
    d3.selectAll('#mean').remove();
  }

  // With these behaviors captured in functions, we just need to hook them up
  // as event listeners on our mark selections!
  countryMarks
    .on('mouseover', function (event, data) {
      highlightMarks(this.id);
      popUpTooltip(event, data);
    })
    .on('mousemove', function (event, data) {
      adjustTooltip(event);
    })
    .on('mouseleave', function (event, data) {
      resetMarks(this.id);
      hideTooltip();
    });

  scatterMarks
    .on('mouseover', function (event, data) {
      highlightContinent(data.Continent);
      highlightMarks(this.id);
      addCentroid(data.Continent);
    })
    .on('mouseleave', function (event, data) {
      unHighlightContinent(data.Continent);
      resetMarks(this.id);
      removeCentroid();
    });

  d3.select('#yAxisG').on('click', (event, data) => {
    [yScaleType, yAxis] = toggleAxisScale(
      yScaleType,
      yExtent,
      yAttr,
      'y',
    );
  });
}

//////////////////////////////////////////////////////
//
//   NEW AND EXCITING! FRESH OFF THE PRESSES!
//   Here are callback functions that I use to adjust
//   the plots based on newly updated attributes
//
//////////////////////////////////////////////////////

function renderNewXAttr(attrName) {
  xAttr = attrName;
  xExtent = getDataExtent(allCountryData.values(), xAttr);
  xScale.domain(xExtent);
  redrawXAxis(1000, xScale, xAttr);
  d3.selectAll('circle.dot')
    .transition()
    .duration(2000)
    .attr('cx', (d) => xScale(d[xAttr]));
}

function renderNewYAttr(attrName) {
  yAttr = attrName;
  yExtent = getDataExtent(allCountryData.values(), yAttr);
  yScale.domain(yExtent);
  redrawYAxis(1000, yScale, yAttr);
  d3.selectAll('circle.dot')
    .transition()
    .duration(2000)
    .attr('cy', (d) => yScale(d[yAttr]));
}

function renderNewMapColor(attrName) {
  console.log(attrName);
  mapColorAttr = attrName;
  mapColorScale = d3
    .scaleLinear()
    .domain(
      getDataExtent(allCountryData.values(), attrName),
    )
    .range(['white', 'blue']);
  console.log(mapColorScale.domain());
  d3.selectAll('path.poly').attr('fill', getCountryColor);
}

//////////////////////////////////////////////////////
//
//  Here is where I hook the plot adjustment callbacks
//  up to my <input> tags that I've already created in
//  index.html
//
//////////////////////////////////////////////////////

function setControls(countryData) {
  // I can use the "id" attribute in index.html to make it easy
  // to find specific elements that I want to add interactivity to

  // First, the axis toggle buttons:
  d3.select('#yAxisToggle').on('click', () => {
    [yScaleType, yAxis] = toggleAxisScale(
      yScaleType,
      yExtent,
      yAttr,
      'y',
    );
  });

  d3.select('#xAxisToggle').on('click', () => {
    [xScaleType, xAxis] = toggleAxisScale(
      xScaleType,
      xExtent,
      xAttr,
      'x',
    );
  });

  // Then, my three dropdown menus
  // The tricky thing here is that I don't want to hard-code
  // the possible menu options (plus, typing out all the strings...
  // yuck!), so I want to just take all the columns of my dataset
  // (a.k.a. the attributes of each row object), and create a new
  // <option> tag for each of them

  // I'm pretty sure Spain has complete data, so lets grab all the
  // attributes of the data object I have stored for Spain
  let validColumns = Object.keys(countryData.get('spain'));

  // Since I'm going to want to add these options to all three
  // dropdown menus, lets loop over the three menu id's and add the
  // option elements to each of them in sequence
  ['xAttrSelect', 'yAttrSelect', 'mapColorSelect'].forEach(
    (id) => {
      d3.select(`#${id}`)
        .selectAll(['empty'])
        // Here's the magic: for each datapoint (my columns), join
        // it to a new HTML element...
        .data(validColumns)
        // ...and I want that HTML element to be an <option>
        .join('option')
        .attr('value', (d) => d)
        .text((d) => d);
    },
  );

  d3.select('#xAttrSelect').on('change', function () {
    renderNewXAttr(d3.select(this).property('value'));
  });

  d3.select('#yAttrSelect').on('change', function () {
    renderNewYAttr(d3.select(this).property('value'));
  });

  d3.select('#mapColorSelect').on('change', function () {
    renderNewMapColor(d3.select(this).property('value'));
  });
}

//Read the data
const mapUrl =
  'https://raw.githubusercontent.com/holtzy/D3-graph-gallery/master/DATA/world.geojson';
const ausUrl =
  'https://raw.githubusercontent.com/tonywr71/GeoJson-Data/master/australian-suburbs.geojson';
const countryStatsUrl =
  'https://raw.githubusercontent.com/bepnye/CP341/main/world-data-2023.csv';
const continentUrl =
  'https://raw.githubusercontent.com/samayo/country-json/master/src/country-by-continent.json';

Promise.all([
  d3.json(mapUrl),
  d3.csv(countryStatsUrl),
  d3.json(continentUrl),
]).then(([mapData, csvData, continentData]) => {
  allCountryData = parseCountryData(csvData, continentData);
  allMapData = parseMapData(mapData);
  allMapData = mapData;
  setControls(allCountryData);
  render(allMapData, allCountryData);
});
